;;; -*- Mode: Lisp; Base: 10; Package: RMS -*-

(herald rms)

;;;; Random macros and utility functions

(eval-when (compile eval load)
(defmacro rmsop (fields &body operations)
  `(si:without-gc-relocation
     ,@(loop for (field var val) in fields
	     collect `(setf (,field ,var) (si:address-of ,val)))
     ,@operations))

(or (fboundp 'si:without-gc-relocation)
    (defmacro si:without-gc-relocation (&body body) `(progn ,@body)))
)

(comment
(defun check-options (operation object options allowed-options)
  (loop for key being the vector-elements of options by 2 using (index i)
	as val = (vref options (1+& i))
	as temp = (assq key allowed-options)
	do (if (null temp)
	       (ferror () "~S unknown option while doing ~S to ~S"
		       key operation object)
	       (unless (or (null (cdr temp))
			   (member val (cdr temp)))
		       (ferror () "~S bad value for option ~S while doing ~S to ~S"
			       val key operation object)))))

(defun find-option (option vector &optional (default nil def-p))
  (loop for k being the vector-elements of vector by 2 using (index i)
	when (if (pairp option)
		 (memq k option)
		 (eq k option))
	 return (values (vref vector (1+& i)) k)
	finally (return (values default def-p))))
)

(defun losecheck (operation stat no-error)
  (cond (no-error (values (not (si:rms-losing stat)) stat))
	(t (values (not (si:loschk operation stat)) stat))))

;;;; Necessary RMS Constants

;;(defconstant si:rms$_eof
;;  #x+1827A)
(defconstant si:rms$_rnf
  #x+182B2)

;;;; Flavor definitions

(defflavor rms-file
	   ((file-name "")
	    (fab (si:make-fab))
	    (nam (si:make-nam))
	    (truename ""))
	   ()
  (:gettable-instance-variables truename fab)
  (:initable-instance-variables file-name))

(defflavor rms-file-stream
	   (rms-file
	    fab
	    (rab (si:make-rab)))
	   ()
  (:gettable-instance-variables rms-file))

;;;; rms-file methods

(defmethod (rms-file :is-open?) ()
  (not (zerop (si:fab$w_ifi fab))))

(defmethod (rms-file :open) (&optional &key name fna
				       (multi-stream-access nil mse-p)
				       (shared nil shr-p)
				       no-error &aux stat)
  (if name (setq fna name))
  (if fna (setq file-name fna))
  (si:using-resources ((si:string-buffer rsa si:nam$c_maxrss)
		       (si:string-buffer esa si:nam$c_maxrss))
    (setf (si:nam$b_rss nam) si:nam$c_maxrss)
    (setf (si:nam$b_ess nam) si:nam$c_maxrss)
    (setf (si:fab$b_fns fab) (string-length file-name))
    (when shr-p (hack-shared-ness shared fab))	; must come before mse-p
    (when mse-p (setf (si:fab$v_mse fab) (if multi-stream-access 1 0)))
    (rmsop ((si:fab$p_nam fab nam)
	    (si:nam$p_rsa nam rsa)
	    (si:nam$p_esa nam esa)
	    (si:fab$p_fna fab file-name))
	   (setq stat (si:rms$open fab)))
    (if (and no-error (si:rms-losing stat))
	(values nil stat)
	(progn (si:loschk "open" stat)
	       (setq truename
		     (si:%string-replace (make-string (si:nam$b_rsl nam))
					 rsa 0 0 (si:nam$b_rsl nam)))
	       (values truename stat)))))

(defun hack-shared-ness (keys fab)
  (setf (si:fab$b_shr fab) 0)
  (cond ((null keys))
	((eq keys t) (hack-shared-ness-1 '(:get :put :upd :del :mse) fab))
	((symbolp keys) (hack-shared-ness-1 (list keys) fab))
	(t (hack-shared-ness-1 keys fab))))

(defconstant *fab-shared-vield-masks*
  '((:put . 1) (:get . 2) (:del . 4) (:upd . 8)
	       (:mse . 16) (:nil . 32) (:upi . 64)))

(defun hack-shared-ness-1 (keys fab)
  (loop with val = 0
	for key in keys
	as temp = (assq key *fab-shared-vield-masks*)
	when (not (null temp))
	  do (setq val (logior& val (cdr temp)))
	finally (setf (si:fab$b_shr fab) val)))

	; should allow options...
(defmethod (rms-file :close) (&optional &key no-error &aux stat)
  (losecheck "close" (setq stat (si:rms$close fab)) no-error))

(defun open-rms-file (name &restv options)
  (let ((new (make-instance 'rms-file)))
    (multiple-value-bind (truename stat)
      (lexpr-send new ':open ':name name options)
      (values (or (not truename) new) truename stat))))    

;;;; RAB macros

(eval-when (compile eval load)

(defvar *rac-alist* '((:key . #.si:rab$c_key)
		      (:seq . #.si:rab$c_seq)
		      (:rfa . #.si:rab$c_rfa)))

(defvar *rab-option-field-alist*
  '((krf . si:rab$b_krf)
    (rac . si:rab$b_rac)
    (rfa . si:rab$w_rfa)
    (kge . si:rab$v_kge)
    (kgt . si:rab$v_kgt)
    (nlk . si:rab$v_nlk)))

(defvar *rab-special-set* '(rac))

(defmacro rab-set (field &optional bool-p (rab-name 'rab))
  (if bool-p
      `(when ,(symbolconc field "-P")
	     (setf (,(cdr (assq field *rab-option-field-alist*))
		    ,rab-name)
		   (if ,field 1 0)))
      (if (memq field *rab-special-set*)
	  `(,(symbolconc "RAB-SET-" field) ,rab-name ,field self)
	  `(when ,field (setf (,(cdr (assq field *rab-option-field-alist*))
			       ,rab-name)
			      ,field)))))

(defmacro rab-set-list (&rest flds)
  `(progn
    ,@(loop for f in flds
	   collect `(rab-set ,(if (pairp f) (car f) f) ,(pairp f)))))

)

(defun rab-set-rac (rab rac self)
  (when rac
	(let ((temp (assq rac *rac-alist*)))
	  (when (null temp)
		(ferror () "~S bad record access type in :FIND for ~S"
			rac self))
	  (setf (si:rab$b_rac rab) (cdr temp)))))


;;;; rms-file-stream methods

(defmethod (rms-file-stream :connect) (file &optional &key (krf 0)
					    no-error)
  (setq rms-file file)
  (setq fab (send rms-file ':fab))
  (setf (si:rab$b_krf rab) krf)
  (rmsop ((si:rab$p_fab rab fab))
    (losecheck "connect" (si:rms$connect rab) no-error)))

(defun make-rms-stream (file &restv options)
  (let ((stream (make-instance 'rms-file-stream)))
    (lexpr-send stream ':connect file options)
    stream))

(defmethod (rms-file-stream :disconnect) (&optional &key no-error)
  (setq rms-file nil fab nil)
  (losecheck "disconnect" (si:rms$disconnect rab) no-error))

(defun check-access-modes (rac rfa-p key-p rab operation self &aux err)
  (case rac
    (nil				; not explicit
     (cond ((and rfa-p key-p)		; trying both?
	    (ferror () "Can't do both Record File Address and Key access~%~
		        in ~S operation on ~S"
		    operation self))
	   (rfa-p (setf (si:rab$b_rac rab) si:rab$c_rfa))	; force rac
	   (key-p (setf (si:rab$b_rac rab) si:rab$c_key))      	; here, too
	   (t (setf (si:rab$b_rac rab) si:rab$c_seq))))	; neither, so sequential
    (:seq (when (or rfa-p key-p) (setq err t)))
    (:rfa (when key-p (setq err t)))
    (:key (when rfa-p (setq err t))))
  (when err
    (ferror () "Conflicting Record Access Mode, ~S, and inputs~%~
	        to ~S operation in ~S"
	    rac operation self)))

;;; common features of find and get should be abstracted (the only difference
;;; is the requirement for a user buffer in get).

;;; Returns two values:
;;;   1 - T or NIL (record found or not)
;;;   2 - RMS status code
;;; Signalls error conditions unless no-error is set

(defmethod (rms-file-stream :find) (&optional &key
					      (key nil key-p)
					      krf rac
					      (rfa nil rfa-p)
					      (kge nil kge-p)
					      (kgt nil kgt-p)
					      (nlk nil nlk-p)
					      no-error)
  (rab-set-list krf rac rfa (kge) (kgt) (nlk))
  (check-access-modes rac rfa-p key-p rab "find" self)
  (let ((access-mode (si:rab$b_rac rab))
	stat)
    (case access-mode
      (#.si:rab$c_rfa (setq stat (si:rms$find rab)))	; rfa already stuffed
      (#.si:rab$c_seq (setq stat (si:rms$find rab)))	; ignore v_lim for now
      (#.si:rab$c_key
       (cond ((null key-p)
	      (ferror () ":KEY must be specified in :FIND on ~S" self)))
       (setf (si:rab$b_ksz rab) (buffer-size key))
       (rmsop ((si:rab$p_kbf rab key))
	      (setq stat (si:rms$find rab)))))
    (if (= stat si:rms$_rnf)
	(values nil stat)
	(losecheck "find" stat no-error))))

;;; returns the read record size (in bytes), nil if EOF, signalls errors.
;;; second value is the RMS return status.

(defmethod (rms-file-stream :get) (&optional &key (ubf nil buf-p)
					     (key nil key-p)
					     krf rac
					     (rfa nil rfa-p)
					     (kge nil kge-p)
					     (kgt nil kgt-p)
					     (nlk nil nlk-p)
					     no-error)
  (rab-set-list krf rac rfa (kge) (kgt) (nlk))
  (check-access-modes rac rfa-p key-p rab "get" self)
  (let ((access-mode (si:rab$b_rac rab))
	stat buffer)
    (if buf-p
	(setq buffer ubf)
	(setq buffer (make-bits (*& (si:fab$w_mrs fab) 8))))
    (setf (si:rab$w_usz rab) (buffer-size buffer))
    (rmsop ((si:rab$p_ubf rab buffer))
      (case access-mode
	(#.si:rab$c_rfa (setq stat (si:rms$get rab)))	; rfa already stuffed
	(#.si:rab$c_seq (setq stat (si:rms$get rab)))	; ignore v_lim for now
	(#.si:rab$c_key
	 (cond ((not key-p)
		(ferror () ":KEY required in :GET on ~S" self)))
	 (setf (si:rab$b_ksz rab) (buffer-size key))
	 (setf (si:rab$p_kbf rab) (si:address-of key))	; already in rmsop
	 (setq stat (si:rms$get rab)))))
    (cond ((= stat si:rms$_eof) (values nil nil stat))
	  ((and no-error (si:rms-losing stat)) (values nil nil stat))
	  (t (progn (si:loschk "get" stat)	; signalls error if necessary
		     (values buffer (si:rab$w_rsz rab) stat))))))

(defmethod (rms-file-stream :rfa0) ()
  (if (zerop (si:rab$w_isi rab))
      (ferror () "Attempt to get RFA from disconnected stream.")
      (si:rab$l_rfa0 rab)))

(defmethod (rms-file-stream :rfa4) ()
  (if (zerop (si:rab$w_isi rab))
      (ferror () "Attempt to get RFA from disconnected stream.")
      (si:rab$w_rfa4 rab)))

;;;; Random Rab functions

;;; returns the buffer's size in bytes.  For bits buffers, rounds down,
;;; if necessary.

(defun buffer-size (buffer)
  (cond ((typep buffer 'bits) (/& (simple-bit-vector-length buffer) 8))
	((typep buffer 'string) (string-length buffer))
	((typep buffer 'vector) (length buffer))
	(t (length buffer))))

(globalize 'buffer-size 'chaos)
