;Thursday  March 3,1983  22:56  -*- Mode:Lisp; Package:SI; Readtable:CL -*-

;;;; Return to Superior, variously

;The a-list of terminal-device mode states which are (by default) set
; in NIL.  We reset them we we quit back to top-level.
(defparameter *nil-terminal-device-states*
  ())


;The a-list of terminal-device mode states which should be in those
; states when we return to our superior.  We save the state, then
; reset these modes when we exit.  (We don't handle proceeding well!
; Maybe that has to interface to the tty-return handler?)
(defparameter *superior-terminal-device-states*
  '((:passall ())))


(defun clear-terminal-device-states (&optional
				      (states *nil-terminal-device-states*))
  (loop for (name value) in states
	when (send terminal-io :device-mode-available? name)
	  do (send terminal-io :set-device-mode name value)))


(defconstant nmsg$_return
  0)

(defconstant nmsg$_delproc
  1)

(defconstant nmsg$_runproc
  2)

(defconstant nmsg$_do_com
  3)

(defconstant nmsg$_proceed
  4)

(defun check-valret-string (x)
  (cond ((null x) nil)
	((and (stringp x) (<=& (string-length x) 256))
	   (if (simple-string-p x) x (string-append x)))
	(t (check-valret-string
	     (cerror t nil :wrong-type-argument
		     "~*~S must be string not more than 256 long"
		     'string x)))))


(defun quit (&optional error-string)
  (setq error-string (check-valret-string error-string))
  (when (eq (send terminal-io :send-if-handles :status) :open)
    (clear-terminal-device-states *superior-terminal-device-states*))
  (si:return-to-superior 0 nmsg$_delproc (or error-string "")))


(defun valret (&optional command-line)
  (exit-from-nil-1 nmsg$_do_com command-line))


(defun exit-from-nil (&optional message)
  (exit-from-nil-1 nmsg$_return message))


(defun exit-and-run-program (pathname)
  (exit-from-nil-1 nmsg$_runproc (send (pathname pathname) :string-for-host)))


(defun proceed-nil (&optional message)
  (setq message (check-valret-string message))
  (when (eq (send terminal-io :send-if-handles :status) :open)
    (clear-terminal-device-states *superior-terminal-device-states*))
  (si:return-to-superior 0 nmsg$_proceed (or message "")))


(defun exit-from-nil-1 (code string &aux states)
  (setq string (check-valret-string string))
  (unwind-protect
    (progn (and (eq (send terminal-io :send-if-handles :status) :open)
		(setq states (send terminal-io :send-if-handles
				   :get-device-mode-states))
		(clear-terminal-device-states
		  *superior-terminal-device-states*))
	   (si:return-to-superior
	     0 (if string code nmsg$_return) (or string "")))
    (when states (send terminal-io :restore-device-mode-states states))))


;;;; start up lisp

(defstruct (start-up-lisp :extend :conc-name)
	   (fp-contents (fp-contents))
	   (ap-contents (ap-contents))
	   (sp-contents (sp-contents)))

(defvar start-up-lisp ()
	"Contains Stack-Info the GC might want to look at")

(defvar lisp-debugger-on-exceptions
  ()
  "Whether or not VMS exceptions enter get Lisp or VMS debugger")

(defvar start-up-lisp-called? ())

(defun start-up-lisp-1 ()
  ;; This takes care of herald b.s. from primitive load.
  (unless (null file-exit-functions)
    (do ((fns file-exit-functions (cdr fns)))
	((atom fns) (setq file-exit-functions ()))
      (funcall (car fns) ())))
  (if start-up-lisp-called?
      (cerror t () ':mumble "Start-Up-Lisp getting called again!")
      ;; The logical name NIL$LINKING will be bound to either
      ;; "NEW" "OLD" or "EXISTING"
      (let ((hack (probe-file "NIL$DISK:[NIL.SITE]SITEPARAM.VAS")))
	(when hack (load hack :set-default-pathname nil :static t)))
      (when (dumpout (or (translate-logical-name "NIL$LINKING") "NEW")
		     (or (translate-logical-name "NIL$NEWCHAOS") "NEW"))
	;DUMPOUT returns what SUSPEND returns:  T if just dumped,
	; () if restored.
	;(print "Linking done. Now ^Y and LXNIL/KILL")
	;Except when it doesn't.
	)))


(defun start-up-lisp ()
  ;; This function gets called after initial linking is done,
  ;; (during which the VASLOADER is being called via a loop
  ;; written in BLISS). At this time the "lisp world" takes over.
  (*catch 'start-up-lisp ;; an experiment I was doing.
    (let ((start-up-lisp (make-start-up-lisp)))
      (*catch 'ferror			;Quit from debugger
	(*catch 'si:top-level		;Hard quit
	  (*catch 'si:errset-catch	;Soft quit
	    (start-up-lisp-1))))
      (setq start-up-lisp-called? T)
      (do () (())
	;Extra level of nesting to allow for timing screws with interrupts
	; before we reestablish the catch frames.
	(*catch 'si:top-level
	  (*catch 'ferror
	    (do () (())
	      (when (eq (send terminal-io :send-if-handles :status) :open)
		(clear-terminal-device-states *NIL-terminal-device-states*))
	      (*catch 'si:top-level
		(*catch 'ferror (repl nil nil t))))))))))

;;;; dumpout

;; When LXNIL is called the logical name NIL$LINKING may be
;; defined to control what happens after linking is done.
;; "NEW" says create a new experimental version of lisp, "OLD"
;; means that the *.VAS files have not "changed," so consider this
;; an "OLD" version which is being relinked for another reason,
;; such as installation on a new site, or after having changed
;; certain SYSGEN parameters. In that case, also load Patches
;; if any.

(DEFUN DUMPOUT (&OPTIONAL (X "NEW")
			  (Y "NEW"))
  (cond ((member x '("NEW" "OLD") :test #'string-equal)
	   (setq lisp-debugger-on-exceptions T)
	   (setq reader-to-use 'aread)
	   (setq *new-printer? T)
	   (settup-lisp-patch-system
	      (string-equal x "NEW") (string-equal x "OLD"))
	   (settup-chaos-patch-system
	      (string-equal y "NEW") (string-equal y "OLD"))
	   (format terminal-io "~&; Suspending Environment~%")
           (prog1 (suspend (or (translate-logical-name "NIL$SAVENAME")
			       "NIL$DISK:[NILCHAOS]CHNIL"))
             (load-fix-files-and-print-herald)
             (print (string-append (si:system-version-info) "\"))))))


(DEFUN SETTUP-CHAOS-PATCH-SYSTEM (INCP LOADP)
  (lexpr-funcall #'si:new-patch-system
		 "Chaos" "NIL$DISK:[NILCHAOS.PATCH]SYSDEF"
		 (if incp () '(:DEFINE)))
  (or incp (AND LOADP (let ((*load-in-static-area* t))
			(load-patches ':noselective)))))

(DEFUN SETTUP-LISP-PATCH-SYSTEM (INCP LOADP)
  (lexpr-funcall #'si:new-patch-system
		 "Lisp" "NIL$DISK:[NIL.PATCH]SYSDEF"
		 (if incp () '(:DEFINE)))
  (or incp (AND LOADP (let ((*load-in-static-area* t))
			(load-patches ':noselective)))))

(defun load-fix-files-and-print-herald ()
  (let ((fix-file-name (format () "LOCAL:NIL$DISK:[NIL.DUMP]FIX~D.LSP"
                               (status lispv))))
    (let ((fix-file-truename (probe-file fix-file-name)))
      (if fix-file-truename
          ;; LOAD is verbose enough.
          (load fix-file-truename :set-default-pathname nil))))
  ;Print herald.  Second arg T means look externally to update statuses
  ; and show user what most recent minor version # is, so he will realize
  ; he should load patches.  Better is for the site initialization file
  ; NIL$DISK:[NIL.SITE]DEFAULT.INI to load patches by default.  Then, the
  ; user can only get away without patches if he has his own.
  (si:print-herald () t)
  (let* ((init-file-name (init-file-pathname "NIL"))
         (homedir (user-homedir-pathname))
         (oinit-file-name (make-pathname
                             ':host (send homedir ':host)
                             ':device (send homedir ':device)
                             ':directory (send homedir ':directory)
                             ':name "NILINIT" ':type "LISP"))
         (name (cond ((probe-file init-file-name) init-file-name)
                     ((probe-file oinit-file-name) oinit-file-name)
                     ((probe-file "NIL$DISK:[NIL.SITE]DEFAULT.INI")
                        "NIL$DISK:[NIL.SITE]DEFAULT.INI"))))
    (when name (load name :set-default-pathname nil))))

;Returns T if it dumped, () if restored.
(defun suspend (lsv-filename
                &optional (system-name "NIL")
                          (system-comment "This is the initial Lisp system"))
  (let ((old si:print-gc-messages) (si:print-gc-messages nil))
    (prog1 (si:suspend-internal lsv-filename system-name system-comment)
      (open primitive-terminal-io)  ;Side effects terminal-io
      (setq si:print-gc-messages old)
      (si:determine-and-set-terminal-type)    ; does canonicalize-cursor-state
      (si:inquire-as-to-environment)
      (fs:initialize-filesystem-variables)
      )
    )
  )
